// KPLOT - A Program for Plotting and Analyzing Crystal Structures 
// Copyright (C) 1979-2019 Rudolf Hundt 
// KPLOT is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
// as published by the Free Software Foundation, version 3.
// KPLOT is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
// See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see https://www.gnu.org/licenses/
// Contact: Rudolf Hundt, University of Bonn Klaus Pleuger,
// Co-Programmer, mailto:k.pleuger@t-online.de

//
// Draws a gear.
//
// Portions of this code have been borrowed from Brian Paul's Mesa
// distribution.
//

#include <qgl.h>
#include <qapplication.h>
#include <math.h>
#include "kplot3d.h"
#include "GL/glut.h"
#include "kplot.h"
#include "kplot2d.h"
#include "qcolor.h"
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <qimage.h>
#include "qkeycode.h"

extern CKplotApp * kplotApp;
static GLfloat colors[16][4];
static    float 			m_vecCamPos[3] ;
static     float 			m_vecCenter[3] ;

#define EPSILON 0.001

#define ABS(x)	((x) >= 0 ? (x) : -(x))
#define	MIN(x,y) ((x) <= (y) ? (x) : (y))
#define	MAX(x,y) ((x) >= (y) ? (x) : (y))
#define SIGN(x) (((x)<0) ? -1 : 1)
#define Max(a,b) (((a)>(b))?(a):(b))

#define EPS_SMOOTH_LINE_FACTOR 0.06  /* Lower for better smooth */

const float CKPlot3D::vecResetCamPos[3] = { 0.0, 0.0, 10.985 } ;
const float CKPlot3D::vecInitCenter[3] = { 0.0, 0.0, 0.0 } ;


void CKPlot3D::draw()
{
    angle += rotateStep;
/*    if (atomList.count()!=0)
    {
       Atom3d *at=atomList.at(0);
       if (at->glDisplayList==0)
           return; // SetupGL();
    }
*/
    static GLfloat pos[4] = {5.0, 5.0, 10.0, 1.0 };
    
    
    pos[0]=m_vecCamPos[0];
    pos[1]=m_vecCamPos[1];
    pos[2]=m_vecCamPos[2];
    

    glLightfv( GL_LIGHT0, GL_POSITION, pos );
    glEnable( GL_CULL_FACE );
    glEnable( GL_LIGHTING );
    glEnable( GL_LIGHT0 );
    glEnable( GL_DEPTH_TEST );
    glEnable( GL_NORMALIZE );
    glShadeModel( GL_SMOOTH );
	if (polygonList.count())
	{
	   glDisable(GL_LIGHTING);
	   glDisable(GL_CULL_FACE);
//	   SetupGL();
	}
    glClearColor(colors[0][0],colors[0][1],colors[0][2],colors[0][3]);
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    glPushMatrix();
    glRotatef( angle/*view_rotx*/, 1.0, 0.0, 0.0 );
    glRotatef( 0/*view_roty*/, 0.0, 1.0, 0.0 );
    glRotatef( 0/*view_rotz*/, 0.0, 0.0, 1.0 );


    for (int i=0;i<atomListDisp.count();i++)
    {
       Atom3dDisp *at=atomListDisp.at(i);
       Atom3d *at2=getAtom(at->nr);
       if (at2==0) {
         continue;}
       glPushMatrix();
       glTranslatef(at2->x,at2->y,at2->z);
       glCallList(at->glDisplayList);
       glPopMatrix();   
    } 

    for (int i=0;i<bondList.count();i++)
    {
       Bond3d *bond=bondList.at(i);
       Atom3d *at=getAtom(bond->p2);
       if (at==0) {
         continue;}
       glPushMatrix();
       glTranslatef(at->x,at->y,at->z);
       glRotatef(bond->yAngle,0,1,0);
       glRotatef(bond->xAngle,1,0,0);
       glRotatef(0,0,0,1);
       glCallList(bond->glDisplayList);
       glPopMatrix();   
    } 

    for (int i=0;i<polygonList.count();i++)
    {
	   Polygon3D *poly=polygonList.at(i);
	   glPushMatrix();
	   glTranslatef(0,0,0);
//	   glColor3f(1,1,1);
	   glCallList(poly->glDisplayList);
	   glPopMatrix();
	}
 
//cout << "Polygone : " << polygonList.count() << endl;    
    glPopMatrix();

}



CKPlot3D::CKPlot3D( QWidget *parent, const char *name )
     : QGLWidget( parent, name )
{
    atomList.setAutoDelete(TRUE);
    bondList.setAutoDelete(TRUE);
    zPos=40;
    zPosOrg=40;
    angle=0;
    zPosDiff=0;
    setColorTable();
    timer=50;
    rotateStep=0;
    segments=6;
    xPosDiff=0;
    yPosDiff=0;


    // Button and key states
    m_enDragMode = DM_NONE ;
    m_shift = 0 ;


    // Current viewport size
    m_xsize = 0 ;
    m_ysize = 0 ;

    m_vecCamPos[0]=0.0;
    m_vecCamPos[1]=0.0;
    m_vecCamPos[2]=10.0;
    m_vecCenter[0]=0.0;
    m_vecCenter[1]=0.0;
    m_vecCenter[2]=0.0;
    setFocusPolicy( QWidget::StrongFocus ) ;
    setFocus() ;
    setCursor( crossCursor ) ;
    glFormat=new QGLFormat();

    img1=new QImage( "marble.gif" );	// Load first image from file
    if ( img1->isNull() ) {		// File not found handling
	warning( "Could not read image file, using single-color instead." );
	QImage dummy( 128, 128, 32 );
	dummy.fill( Qt::red.rgb() );	
	*img1 = dummy;
    }
    else {
	*img1 = img1->convertDepth( 32 );	// Make sure it has the right depth
	*img1 = img1->mirror();		// OpenGL wants it upside down
    }
	
	aktRenderPixmap=0;
 

}


void CKPlot3D::initializeGL()
{
   QGLWidget::initializeGL();
   SetupGL();
}


void CKPlot3D::resizeGL( int width, int height )
{
//	setFormat(glFormat);
    GLfloat w = (float) width / (float) height;
    GLfloat h = 1.0;
    m_xsize=width;
    m_ysize=height;
    glViewport( 0, 0, width, height );
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum( -w, w, -h, h, 5.0, zPos*10.0 );

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
//    glTranslatef( 0.0, 0.0, zPos/2.0*(-1) );
//    glTranslatef( xPosDiff, yPosDiff, (zPos+zPosDiff)*(-1) );
    gluLookAt( m_vecCamPos[0], m_vecCamPos[1], m_vecCamPos[2], 
               m_vecCenter[0], m_vecCenter[1], m_vecCenter[2], 
               0.0, 1.0, 0.0 );
//    cout << m_vecCamPos[0] << " " << m_vecCamPos[1] << " " << m_vecCamPos[2] << endl;
//    glTranslated(-m_vecCamPos[0],-m_vecCamPos[1],-m_vecCamPos[2]);
/*    static GLfloat pos[4] = {5.0, 5.0, 10.0, 1.0 };
    glLightfv( GL_LIGHT0, GL_POSITION, pos );
    glEnable( GL_CULL_FACE );
    glEnable( GL_LIGHTING );
    glEnable( GL_LIGHT0 );
    glEnable( GL_DEPTH_TEST );
*/
}


void CKPlot3D::paintGL()
{
    draw();
}

void CKPlot3D::timerEvent(QTimerEvent*)
{
    updateGL();
}

void CKPlot3D::SetAtom3DDisp(long int nr, float r, long int color) {
//  cout << "ANFA" << endl;
   GLuint w;
   Atom3dDisp *at=new Atom3dDisp();
   at->r=r;
   at->color=((color)%16);
   at->nr=nr;
   at->glDisplayList=0;
   atomListDisp.append(at);
//   cout << "Atom : " << *nr << endl;
}


void CKPlot3D::SetAtom3D(float x, float y,float z, long int nr) {
//  cout << "ANFA" << endl;
   GLuint w;
   Atom3d *at=new Atom3d();
   at->x=x;
   at->y=y;
   at->z=z;
//   at->r=*r;
   at->nr=nr;
//   at->color=((*color)%16);
//   at->nr=*nr;
//   at->glDisplayList=0;
   atomList.append(at);
//   cout << "Atom : " << *nr << endl;
}

void CKPlot3D::ResetDataModel()
{
    for (int i=0;i<atomListDisp.count();i++)
    {
       Atom3dDisp *at=atomListDisp.at(i);
       if (at->glDisplayList!=0)
          glDeleteLists(at->glDisplayList,1);
     } 

    for (int i=0;i<bondList.count();i++)
    {
       Bond3d *bond=bondList.at(i);
       if (bond->glDisplayList!=0)
          glDeleteLists(bond->glDisplayList,1);
    }
    for (int i=0;i<polygonList.count();i++)
    {
       Polygon3D *bond=polygonList.at(i);
       if (bond->glDisplayList!=0)
          glDeleteLists(bond->glDisplayList,1);
    }

    atomList.clear();
    bondList.clear(); 
    polygonList.clear();
    atomListDisp.clear();

}

void CKPlot3D::Reset()
{
//    zPos=40;
    zPosDiff=0;
    xPosDiff=0;
    yPosDiff=0;
    angle=0;
/*    if (bondList.count()!=0)
    {
       Bond3d *at=bondList.at(0);
       zPos=at->zPos;
    }
*/
    zPos=zPosOrg;
    /* Move the viewer back to the starting point */
    int ii ;
    for ( ii = 0 ; ii < 3 ; ii++ )
        m_vecCamPos[ii] = vecResetCamPos[ii] ;
    for ( ii = 0 ; ii < 3 ; ii++ )
        m_vecCenter[ii] = vecInitCenter[ii] ;
    resizeGL(width(),height());
    SetupGL();
    emit updateGL();
}

void CKPlot3D::SetupGL()
{

    static GLfloat pos[4] = {5.0, 5.0, 10.0, 1.0 };
 

    SetRotateAngles();
//    glLightfv( GL_LIGHT0, GL_POSITION, pos );
//    glEnable( GL_CULL_FACE );
//    glEnable( GL_LIGHTING );
//    glEnable( GL_LIGHT0 );
    glEnable( GL_DEPTH_TEST );
if (aktRenderPixmap)
{
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glEnable(GL_TEXTURE_2D);
}

    for (int i=0;i<atomListDisp.count();i++)
    {
       Atom3dDisp *at=atomListDisp.at(i);
       if (at->glDisplayList!=0)
          glDeleteLists(at->glDisplayList,1);
       GLuint w;
       GLUquadricObj *qobj;
       qobj=gluNewQuadric();
       w = glGenLists(1);
       glNewList(w, GL_COMPILE);
       glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, colors[at->color] );
	   glColor3f(colors[at->color][0],colors[at->color][1],colors[at->color][2]);
if (aktRenderPixmap)
{
    glTexImage2D( GL_TEXTURE_2D, 0, 3, img1->width(), img1->height(), 0,
		  GL_RGBA, GL_UNSIGNED_BYTE, img1->bits() );
    gluQuadricNormals(qobj, GL_SMOOTH);
    gluQuadricTexture(qobj, GL_TRUE);
}
       if (solidFlag)
          gluQuadricDrawStyle(qobj,(GLenum)GLU_FILL);
       else
          gluQuadricDrawStyle(qobj,(GLenum)GLU_LINE);
       gluSphere(qobj,at->r,segments,segments);
       glEndList();
       at->glDisplayList=w;

/*       QString str="Hallo";
       char s[100];
       sprintf(s,"%d",atomList.count());
       str+=s;
       kplotApp->output(str);
*/
     }
 

    for (int i=0;i<bondList.count();i++)
    {
       Bond3d *bond=bondList.at(i);
       if (bond->glDisplayList!=0)
          glDeleteLists(bond->glDisplayList,1);
       Atom3d *at=getAtom(bond->p1);
       if (at==0) continue;
       GLuint w;
       w = glGenLists(1);
       GLUquadricObj *qobj;
       qobj=gluNewQuadric();
       glNewList(w, GL_COMPILE);
       glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, colors[bond->color] );
	   glColor3f(colors[bond->color][0],colors[bond->color][1],colors[bond->color][2]);
if (aktRenderPixmap)
{       
    glTexImage2D( GL_TEXTURE_2D, 0, 3, img1->width(), img1->height(), 0,
		  GL_RGBA, GL_UNSIGNED_BYTE, img1->bits() );
    gluQuadricNormals(qobj, GL_SMOOTH);
    gluQuadricTexture(qobj, GL_TRUE);
}
       if (solidFlag)
          gluQuadricDrawStyle(qobj,(GLenum)GLU_FILL);
       else
          gluQuadricDrawStyle(qobj,(GLenum)GLU_LINE);
       gluCylinder(qobj,bond->dm,bond->dm,bond->length,segments,segments);
       glEndList();
       bond->glDisplayList=w;
     } 

    for (int i=0;i<polygonList.count();i++)
    {
       Polygon3D *poly=polygonList.at(i);
       if (poly->glDisplayList!=0)
          glDeleteLists(poly->glDisplayList,1);
       GLuint w;
       w = glGenLists(1);
       glNewList(w, GL_COMPILE);
/*      glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, colors[poly->color] );
if (aktRenderPixmap)
{       
    glTexImage2D( GL_TEXTURE_2D, 0, 3, img1->width(), img1->height(), 0,
		  GL_RGBA, GL_UNSIGNED_BYTE, img1->bits() );
}
*/
/*       if (solidFlag)
          gluQuadricDrawStyle(qobj,GLU_FILL);
       else
          gluQuadricDrawStyle(qobj,GLU_LINE);
       gluCylinder(qobj,bond->dm,bond->dm,bond->length,segments,segments);*/
       if (solidFlag)
       {
       glBegin(GL_POLYGON);
       glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
       glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, colors[poly->color] );
	   glColor3f(colors[poly->color][0],colors[poly->color][1],colors[poly->color][2]);
	   for(int x=0;x<poly->atome.count();x++)
	   {
       Atom3d *at=getAtom(*(poly->atome.at(x)));
       if (at==0) continue;
       glVertex3f(at->x,at->y,at->z);
       }
       glEnd();
       }

       {
       glBegin(GL_LINE_LOOP);
       glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, colors[poly->muster] );
	   glColor3f(colors[poly->muster][0],colors[poly->muster][1],colors[poly->muster][2]);
	   for(int x=0;x<poly->atome.count();x++)
	   {
//      glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, colors[poly->muster] );
       Atom3d *at=getAtom(*(poly->atome.at(x)));
       if (at==0) continue;
       glVertex3f(at->x,at->y,at->z);
       }
       glEnd();
       }
       glEndList();
       poly->glDisplayList=w;
     } 

    glEnable( GL_NORMALIZE );
    glShadeModel( GL_SMOOTH );
   
}

void CKPlot3D::Animate(int t)
{
    timer=t;;
}

void CKPlot3D::SetBond3D(long int  p1, long int  p2, long int 
color, float  radius)
{
//   if (atomList.count()<*p2)
//  cout << "ANFB" << endl;
   {
     Bond3d *at=new Bond3d();
     at->p1=p1;
     at->p2=p2;
     at->color=((color)%16);
     at->xAngle=0;
     at->yAngle=0;
     at->length=0;
     at->dm=radius;
     at->glDisplayList=0;
     bondList.append(at);
/*
     QString str="";
     char s[100];
     sprintf(s,"Save P1 : %d , P2 : %d",*p1,*p2);
     str+=s;
//     cout << str << endl;
     kplotApp->output(str);
*/
   }
}

Atom3d * CKPlot3D::getAtom(int nr)
{
    int t=0;
    Atom3d * ret=0;
    for (t=0;t<atomList.count();t++)
    {
       ret=atomList.at(t);
       if (ret->nr==nr)
           break; 
    }
    if (t>=atomList.count())
    {
        ret=0;
    }
    return ret;
}
     
void CKPlot3D::SetRotateAngles()
{
    static int id2[3]={1,2,0};
    float q[3],u[3];
    float cosf,sinf;

    for (int t=0;t<bondList.count();t++)
    {
        Bond3d *bond=bondList.at(t);
        Atom3d p;
        Atom3d *p1=getAtom(bond->p1); 
        Atom3d *p2=getAtom(bond->p2);
        if(p1==0 || p2==0)
        {
           continue;
        }
//        cout << "Bond " << bond->p1 << " " << bond->p2 << " " 
//          << atomList.count() << endl;
        p.x=p1->x-p2->x;
        p.y=p1->y-p2->y;
        p.z=p1->z-p2->z;

//        p.x=3;
//       p.y=3;
//        p.z=0;
        bond->length=sqrt(p.x*p.x+p.y*p.y+p.z*p.z);
        float d1=sqrt(p.x*p.x+p.z*p.z);
        float f2=0,f1=0;
        float temp;
        float a[3][3];
        float bog=57.29577951;
        int i,j,k;
        f2=0;
        if (fabs(d1)>1.e-10)
        {
           cosf=p.z/d1;
           sinf=p.x/d1;
           temp=cosf*cosf+sinf*sinf;
           f2=acos(cosf)*bog;
           if (sinf<0) f2=-f2;
           for (i=0;i<3;i++)
             for (j=0;j<3;j++)
                a[i][j]=0;
           i=1;
           j=id2[i];
           k=id2[j];
           a[i][i]=1;
           a[j][j]=cosf;
           a[k][k]=cosf;
           a[j][k]=sinf;
           a[k][j]=-sinf;
           for (i=0;i<3;i++)
             q[i]=a[i][0]*p.x+a[i][1]*p.y+a[i][2]*p.z;
         }
         else
         {
           f2=0;
           q[0]=0;
           q[1]=0;
           q[2]=bond->length;
         }

           cosf=q[2]/bond->length;
           sinf=q[1]/bond->length;
           if(cosf>=1.0)
           {
              f1=0;
           } else if(cosf<=-1.0)
           {
              f1=180.0;
	   }
           else
           {
              f1=-acos(cosf)*bog;
           }
           if (sinf < 0) f1=-f1;
           for (i=0;i<3;i++)
             for (j=0;j<3;j++)
                a[i][j]=0;
           i=0;
           j=id2[i];
           k=id2[j];             
           a[i][i]=1;
           a[j][j]=cosf;
           a[k][k]=cosf;
           a[j][k]=-sinf;
           a[k][j]=sinf;
           for(i=0;i<3;i++)
              u[i]=a[i][0]*q[0]+a[i][1]*q[1]+a[i][2]*q[2];
           bond->xAngle=f1;
           bond->yAngle=f2;

/*
     QString str="";
     char s[200];

     sprintf(s,"Save P1 : %d , P2 : %d   X Angle : %f, Y Angle : %f"
           ,bond->p1,bond->p2,bond->xAngle,bond->yAngle );
      str+=s;
     kplotApp->output(str);
                 
*/        
    }     
}

void CKPlot3D::BeginDraw()
{
 ResetDataModel();
}

void CKPlot3D::EndDraw()
{
//cout << "ENDDRAW" << endl;
    resizeGL(width(),height());
    SetupGL();
    emit updateGL();
}

void CKPlot3D::outputPS(QString str)
{
    resizeGL(width(),height());
    SetupGL();
    updateGL();
    outputEPS(4000000, 1, (char*)(const char*)str);
}

void CKPlot3D::setColorTable()
{
    for (int i=0;i<16;i++)
    {
//        cout << i << endl;
        QColor *c=CKPlot2D::color[i];
        QRgb rgb=c->rgb();
//        cout << rgb << endl;
        colors[i][2]=(GLfloat)(rgb&255)/255.0;
        colors[i][1]=(GLfloat)((rgb >> 8)&255)/255.0;
        colors[i][0]=(GLfloat)((rgb >> 16)&255)/255.0;
        colors[i][3]=(GLfloat)1.0;
//        cout << colors[i][0] << colors[i][1] << colors[i][2] << endl;
    }
}

void CKPlot3D::setRotate3D(bool flag)
{
   static int id=0;
   if (flag==TRUE)
   {
     rotateStep=1.0;
     id=startTimer(timer);
   }
   else
   {
     rotateStep=0;
     killTimer(id);
   }
}

void CKPlot3D::changeBackgroundColor()
{
   GLfloat c1=colors[15][0];
   GLfloat c2=colors[15][1];
   GLfloat c3=colors[15][2];
   GLfloat c4=colors[15][3];

   colors[15][0]=colors[0][0];
   colors[15][1]=colors[0][1];
   colors[15][2]=colors[0][2];
   colors[15][3]=colors[0][3];

   colors[0][0]=c1;
   colors[0][1]=c2;
   colors[0][2]=c3;
   colors[0][3]=c4;
   	
//   cout << "Change color" << endl;
}

void CKPlot3D::setSolidMode(bool flag)
{
   solidFlag=flag;
}

void CKPlot3D::setRenderMode(bool flag)
{
   if (flag)
      segments=16;
   else
      segments=6;
}

void CKPlot3D::incrementZPos() 
{
   zPosDiff+=0.5;
   resizeGL(width(),height());
   emit updateGL();
}

void CKPlot3D::decrementZPos()
{
   zPosDiff-=0.5;
   resizeGL(width(),height());
   emit updateGL();
}

void CKPlot3D::setBufferedMode(bool flag)
{
   glFormat->setDoubleBuffer(flag);
   setFormat(*glFormat);
   makeCurrent();
   glInit();
   resizeGL(width(),height());
   emit updateGL();
}

/*
void CKPlot3D::mouseMoveEvent(QMouseEvent * e)
{
//     cout << "Mouse" << endl;
     static int x=0,y=0;


     if (e->state()==RightButton)
     {

         if (e->x()>x)
            xPosDiff+=1;
         else if (e->x()<x)
            xPosDiff-=1;
         if (e->y()>y)
            yPosDiff-=1;
         else if(e->y()<y)
            yPosDiff+=1;
         resizeGL(width(),height());
         emit updateGL();
     }
     x=e->x();
     y=e->y();

}
*/

void CKPlot3D::mouseMoveEvent( QMouseEvent * pME )
{
    float newxobs, newyobs, newzobs;
    float newxcenter, newycenter, newzcenter;

    if ( m_enDragMode != DM_NONE ) 
    {
        float mx, my ;
        float dx, dy ;
        /* Only do mouse motion when the buttons are down! */
        mx = ((float) pME->x())/(m_xsize);
        my = ((float) pME->y())/(m_ysize);
        dx = 2*(mx - m_lx);
        dy = -2*(my - m_ly);
        m_lx = mx;
        m_ly = my;

        if ( m_enDragMode == DM_ZOOM ) 
        {
            /* Mouse Y maps to zooming toward/from center
             * mouse X does nothing
             * Movement slows down as you approach the far plane
             * Don't let observer reach far plane unless multifield
             */
            if (dy < 0) 
            {
                m_vecCamPos[0] = m_vecCamPos[0]+(m_vecCamPos[0]-m_vecCenter[0])*dy;
                m_vecCamPos[1] = m_vecCamPos[1]+(m_vecCamPos[1]-m_vecCenter[1])*dy;
                m_vecCamPos[2] = m_vecCamPos[2]+(m_vecCamPos[2]-m_vecCenter[2])*dy;
            }
            else 
            {
                m_vecCamPos[0] = m_vecCenter[0]+(m_vecCamPos[0]-m_vecCenter[0])/(1-dy);
                m_vecCamPos[1] = m_vecCenter[1]+(m_vecCamPos[1]-m_vecCenter[1])/(1-dy);
                m_vecCamPos[2] = m_vecCenter[2]+(m_vecCamPos[2]-m_vecCenter[2])/(1-dy);
            }
            /* Restore dy to original sign */
            dy = -dy;
        }
        else if ( m_enDragMode == DM_ORBIT ) 
        {
            /* Mouse XY maps to longitude and lattitude
             * as observer orbits center at fixed distance
             * If shifted, suppress lattitude changes,
             * easing flyarounds of full-surround fields
             */
            float aold,anew;
            float rold,rnew ;
            rold = sqrt((m_vecCamPos[0]-m_vecCenter[0])*(m_vecCamPos[0]-m_vecCenter[0])+
                        (m_vecCamPos[2]-m_vecCenter[2])*(m_vecCamPos[2]-m_vecCenter[2]));
            aold = atan((m_vecCamPos[0]-m_vecCenter[0])/(m_vecCamPos[2]-m_vecCenter[2])) / M_PI*180.0;
            if (ABS(m_vecCamPos[2]-m_vecCenter[2]) < EPSILON) aold = 90.0 * 
                                                 SIGN(m_vecCamPos[0]-m_vecCenter[0]);
            else if (m_vecCamPos[2]<m_vecCenter[2]) aold = 180.0 + aold;
            anew = aold - dx * 90.0;

            m_vecCamPos[0] = m_vecCenter[0] + rold * sin(anew/180.0*M_PI);
            m_vecCamPos[2] = m_vecCenter[2] + rold * cos(anew/180.0*M_PI);

            rold = sqrt((m_vecCamPos[0]-m_vecCenter[0])*(m_vecCamPos[0]-m_vecCenter[0])+
                        (m_vecCamPos[1]-m_vecCenter[1])*(m_vecCamPos[1]-m_vecCenter[1])+
                        (m_vecCamPos[2]-m_vecCenter[2])*(m_vecCamPos[2]-m_vecCenter[2]));
            newyobs = m_vecCamPos[1]-dy*rold;
            if (!m_shift) 
            {
                m_vecCamPos[1] = newyobs;
            }
            rnew = sqrt((m_vecCamPos[0]-m_vecCenter[0])*(m_vecCamPos[0]-m_vecCenter[0])+
                        (m_vecCamPos[1]-m_vecCenter[1])*(m_vecCamPos[1]-m_vecCenter[1])+
                        (m_vecCamPos[2]-m_vecCenter[2])*(m_vecCamPos[2]-m_vecCenter[2]));
            newxobs = (m_vecCamPos[0]-m_vecCenter[0])*(rold/rnew)+m_vecCenter[0];
            newyobs = (m_vecCamPos[1]-m_vecCenter[1])*(rold/rnew)+m_vecCenter[1];
            newzobs = (m_vecCamPos[2]-m_vecCenter[2])*(rold/rnew)+m_vecCenter[2];
            m_vecCamPos[0] = newxobs;
            if (!m_shift) 
            {
                m_vecCamPos[1] = newyobs;
            }
            m_vecCamPos[2] = newzobs;
        }
        else if ( m_enDragMode == DM_PAN )
        {
            /* Mouse XY maps to translation of viewed image      */
            /* achieved by matched moves of observer and center  */
            /* perpendicular to the ray connecting them          */
            /* Such translation shows off horizontal parallax   */
            if (ABS(m_vecCenter[2]-m_vecCamPos[2]) > ABS(m_vecCenter[0]-m_vecCamPos[0])) 
            {
                /* Front and back views */
                newxcenter = m_vecCenter[0]+(m_vecCenter[2]-m_vecCamPos[2])*dx/2;
                newzcenter = m_vecCenter[2]-(m_vecCenter[0]-m_vecCamPos[0])*dx/2;
                newycenter = m_vecCenter[1]-ABS(m_vecCenter[2]-m_vecCamPos[2])*dy/2;
                newzcenter = newzcenter-(m_vecCenter[1]-m_vecCamPos[1])*dy/2;
                newxobs = m_vecCamPos[0]+(m_vecCenter[2]-m_vecCamPos[2])*dx/2;
                newzobs = m_vecCamPos[2]-(m_vecCenter[0]-m_vecCamPos[0])*dx/2;
                newyobs = m_vecCamPos[1]-ABS(m_vecCenter[2]-m_vecCamPos[2])*dy/2;
                newzobs = newzobs-(m_vecCenter[1]-m_vecCamPos[1])*dy/2;
            }
            else 
            {
                /* Side views */
                newzcenter = m_vecCenter[2]-(m_vecCenter[0]-m_vecCamPos[0])*dx/2;
                newxcenter = m_vecCenter[0]+(m_vecCenter[2]-m_vecCamPos[2])*dx/2;
                newycenter = m_vecCenter[1]-ABS(m_vecCenter[0]-m_vecCamPos[0])*dy/2;
                newxcenter = newxcenter-(m_vecCenter[1]-m_vecCamPos[1])*dy/2;
                newzobs = m_vecCamPos[2]-(m_vecCenter[0]-m_vecCamPos[0])*dx/2;
                newxobs = m_vecCamPos[0]+(m_vecCenter[2]-m_vecCamPos[2])*dx/2;
                newyobs = m_vecCamPos[1]-ABS(m_vecCenter[0]-m_vecCamPos[0])*dy/2;
                newxobs = newxobs-(m_vecCenter[1]-m_vecCamPos[1])*dy/2;
            }
            m_vecCenter[0] = newxcenter;
            m_vecCenter[1] = newycenter;
            m_vecCenter[2] = newzcenter;
            /* If m_shifted, movement of observer is suppressed, */
            /* causing mouse XY to map to viewing direction at */
            /* the (fixed) current position, which exhibits    */
            /* no horizontal parallax but is often useful      */
            if (!m_shift) 
            {
                m_vecCamPos[0] = newxobs;
                m_vecCamPos[1] = newyobs;
                m_vecCamPos[2] = newzobs;
            }
        }

        //loadInteractiveCameraParams() ;

        // Update if we're dragging something
        //update() ;
        resizeGL(width(),height());
        emit updateGL();
    }
}

void CKPlot3D::mousePressEvent( QMouseEvent * pME )
{
    /* Save current mouse position, which was tied to the mouse button. */
    m_lx = ((float) pME->x())/(m_xsize) ;
    m_ly = ((float) pME->y())/(m_ysize) ;

    switch ( pME->button() )
    {
        case LeftButton:
        m_enDragMode = DM_ORBIT ;
        break;

        case MidButton:
        m_enDragMode = DM_PAN ;
        break;

        case RightButton:
        m_enDragMode = DM_ZOOM ;
        break ;
    }

//    update() ;
}

///////////////////////////////////////////////////////////////////////////////
// QViewport3D::mouseReleaseEvent
//
void CKPlot3D::mouseReleaseEvent( QMouseEvent * /*pME*/ )
{
    /* Save current mouse position, which was tied to the mouse button. */
    // m_lx = ((float) pME->x())/(m_xsize) ;
    // m_ly = ((float) pME->y())/(m_ysize) ;


    // Reset drag mode
    //
    // Note: This has to be done AFTER calling the app's handler, because it
    // uses the old value of dragmode to make sure we were dragging something.
    // See: LfViewApp::viewportMouseReleaseEvent.
    m_enDragMode = DM_NONE ;
}

void CKPlot3D::keyPressEvent( QKeyEvent * pKE )
{
    bool fRedraw = true ;

    switch ( pKE->key() )
    {
        case Key_R:
        Reset();
        break;

/*        case Key_Up:
        m_fov *= 1.1;
        loadInteractiveCameraParams() ;
        break;
        case Key_Down:
        m_fov /= 1.1;
        loadInteractiveCameraParams() ;
        break;
*/      
        case Key_Shift:
        fRedraw = false ;
        m_shift = 1 ;
        break;
    }

    // Emit signal for other keys
    emit keyPressed( pKE ) ;

    if ( fRedraw == true )
    {
        update() ;
    }
}

void CKPlot3D::keyReleaseEvent( QKeyEvent * pKE )
{
    switch ( pKE->key() )
    {
        case Key_Shift:
        m_shift = 0 ; 
        break ;
    }
}

/* Write contents of one vertex to stdout. */
void
CKPlot3D::print3DcolorVertex(GLint size, GLint * count,
  GLfloat * buffer)
{
  int i;

  printf("  ");
  for (i = 0; i < 7; i++) {
    printf("%4.2f ", buffer[size - (*count)]);
    *count = *count - 1;
  }
  printf("\n");
}

void
CKPlot3D::printBuffer(GLint size, GLfloat * buffer)
{
  GLint count;
  int token, nvertices;

  count = size;
  while (count) {
    token = buffer[size - count];
    count--;
    switch (token) {
    case GL_PASS_THROUGH_TOKEN:
      printf("GL_PASS_THROUGH_TOKEN\n");
      printf("  %4.2f\n", buffer[size - count]);
      count--;
      break;
    case GL_POINT_TOKEN:
      printf("GL_POINT_TOKEN\n");
      print3DcolorVertex(size, &count, buffer);
      break;
    case GL_LINE_TOKEN:
      printf("GL_LINE_TOKEN\n");
      print3DcolorVertex(size, &count, buffer);
      print3DcolorVertex(size, &count, buffer);
      break;
    case GL_LINE_RESET_TOKEN:
      printf("GL_LINE_RESET_TOKEN\n");
      print3DcolorVertex(size, &count, buffer);
      print3DcolorVertex(size, &count, buffer);
      break;
    case GL_POLYGON_TOKEN:
      printf("GL_POLYGON_TOKEN\n");
      nvertices = buffer[size - count];
      count--;
      for (; nvertices > 0; nvertices--) {
        print3DcolorVertex(size, &count, buffer);
      }
    }
  }
}


static char *gouraudtriangleEPS[] =
{
  "/bd{bind def}bind def /triangle { aload pop   setrgbcolor  aload pop 5 3",
  "roll 4 2 roll 3 2 roll exch moveto lineto lineto closepath fill } bd",
  "/computediff1 { 2 copy sub abs threshold ge {pop pop pop true} { exch 2",
  "index sub abs threshold ge { pop pop true} { sub abs threshold ge } ifelse",
  "} ifelse } bd /computediff3 { 3 copy 0 get 3 1 roll 0 get 3 1 roll 0 get",
  "computediff1 {true} { 3 copy 1 get 3 1 roll 1 get 3 1 roll 1 get",
  "computediff1 {true} { 3 copy 2 get 3 1 roll  2 get 3 1 roll 2 get",
  "computediff1 } ifelse } ifelse } bd /middlecolor { aload pop 4 -1 roll",
  "aload pop 4 -1 roll add 2 div 5 1 roll 3 -1 roll add 2 div 3 1 roll add 2",
  "div 3 1 roll exch 3 array astore } bd /gouraudtriangle { computediff3 { 4",
  "-1 roll aload 7 1 roll 6 -1 roll pop 3 -1 roll pop add 2 div 3 1 roll add",
  "2 div exch 3 -1 roll aload 7 1 roll exch pop 4 -1 roll pop add 2 div 3 1",
  "roll add 2 div exch 3 -1 roll aload 7 1 roll pop 3 -1 roll pop add 2 div 3",
  "1 roll add 2 div exch 7 3 roll 10 -3 roll dup 3 index middlecolor 4 1 roll",
  "2 copy middlecolor 4 1 roll 3 copy pop middlecolor 4 1 roll 13 -1 roll",
  "aload pop 17 index 6 index 15 index 19 index 6 index 17 index 6 array",
  "astore 10 index 10 index 14 index gouraudtriangle 17 index 5 index 17",
  "index 19 index 5 index 19 index 6 array astore 10 index 9 index 13 index",
  "gouraudtriangle 13 index 16 index 5 index 15 index 18 index 5 index 6",
  "array astore 12 index 12 index 9 index gouraudtriangle 17 index 16 index",
  "15 index 19 index 18 index 17 index 6 array astore 10 index 12 index 14",
  "index gouraudtriangle 18 {pop} repeat } { aload pop 5 3 roll aload pop 7 3",
  "roll aload pop 9 3 roll 4 index 6 index 4 index add add 3 div 10 1 roll 7",
  "index 5 index 3 index add add 3 div 10 1 roll 6 index 4 index 2 index add",
  "add 3 div 10 1 roll 9 {pop} repeat 3 array astore triangle } ifelse } bd",
  NULL
};

GLfloat *
CKPlot3D::spewPrimitiveEPS(FILE * file, GLfloat * loc)
{
  int token;
  int nvertices, i;
  GLfloat red, green, blue;
  int smooth;
  GLfloat dx, dy, dr, dg, db, absR, absG, absB, colormax;
  int steps;
  Feedback3Dcolor *vertex;
  GLfloat xstep, ystep, rstep, gstep, bstep;
  GLfloat xnext, ynext, rnext, gnext, bnext, distance;

  token = *loc;
  loc++;
  switch (token) {
  case GL_LINE_RESET_TOKEN:
  case GL_LINE_TOKEN:
    vertex = (Feedback3Dcolor *) loc;

    dr = vertex[1].red - vertex[0].red;
    dg = vertex[1].green - vertex[0].green;
    db = vertex[1].blue - vertex[0].blue;

    if (dr != 0 || dg != 0 || db != 0) {
      /* Smooth shaded line. */
      dx = vertex[1].x - vertex[0].x;
      dy = vertex[1].y - vertex[0].y;
      distance = sqrt(dx * dx + dy * dy);

      absR = fabs(dr);
      absG = fabs(dg);
      absB = fabs(db);


      colormax = Max(absR, Max(absG, absB));
      steps = (int)Max(1.0,colormax*distance*EPS_SMOOTH_LINE_FACTOR); 

      xstep = dx / steps;
      ystep = dy / steps;

      rstep = dr / steps;
      gstep = dg / steps;
      bstep = db / steps;

      xnext = vertex[0].x;
      ynext = vertex[0].y;
      rnext = vertex[0].red;
      gnext = vertex[0].green;
      bnext = vertex[0].blue;

      /* Back up half a step; we want the end points to be
         exactly the their endpoint colors. */
      xnext -= xstep / 2.0;
      ynext -= ystep / 2.0;
      rnext -= rstep / 2.0;
      gnext -= gstep / 2.0;
      bnext -= bstep / 2.0;
    } else {
      /* Single color line. */
      steps = 0;
    }

    fprintf(file, "%g %g %g setrgbcolor\n",
      vertex[0].red, vertex[0].green, vertex[0].blue);
    fprintf(file, "%g %g moveto\n", vertex[0].x, vertex[0].y);

    for (i = 0; i < steps; i++) {
      xnext += xstep;
      ynext += ystep;
      rnext += rstep;
      gnext += gstep;
      bnext += bstep;
      fprintf(file, "%g %g lineto stroke\n", xnext, ynext);
      fprintf(file, "%g %g %g setrgbcolor\n", rnext, gnext, bnext);
      fprintf(file, "%g %g moveto\n", xnext, ynext);
    }
    fprintf(file, "%g %g lineto stroke\n", vertex[1].x, vertex[1].y);

    loc += 14;          /* Each vertex element in the feedback
                           buffer is 7 GLfloats. */

    break;
  case GL_POLYGON_TOKEN:
    nvertices = *loc;
    loc++;

    vertex = (Feedback3Dcolor *) loc;

    if (nvertices > 0) {
      red = vertex[0].red;
      green = vertex[0].green;
      blue = vertex[0].blue;
      smooth = 0;
      for (i = 1; i < nvertices; i++) {
        if (red != vertex[i].red || green != vertex[i].green || blue != vertex[i].blue) {
          smooth = 1;
          break;
        }
      }
      if (smooth) {
        /* Smooth shaded polygon; varying colors at vetices. */
        int triOffset;

        /* Break polygon into "nvertices-2" triangle fans. */
        for (i = 0; i < nvertices - 2; i++) {
          triOffset = i * 7;
          fprintf(file, "[%g %g %g %g %g %g]",
            vertex[0].x, vertex[i + 1].x, vertex[i + 2].x,
            vertex[0].y, vertex[i + 1].y, vertex[i + 2].y);
          fprintf(file, " [%g %g %g] [%g %g %g] [%g %g %g] gouraudtriangle\n",
            vertex[0].red, vertex[0].green, vertex[0].blue,
            vertex[i + 1].red, vertex[i + 1].green, vertex[i + 1].blue,
            vertex[i + 2].red, vertex[i + 2].green, vertex[i + 2].blue);
        }
      } else {
        /* Flat shaded polygon; all vertex colors the same. */
        fprintf(file, "newpath\n");
        fprintf(file, "%g %g %g setrgbcolor\n", red, green, blue);

        /* Draw a filled triangle. */
        fprintf(file, "%g %g moveto\n", vertex[0].x, vertex[0].y);
        for (i = 1; i < nvertices; i++) {
          fprintf(file, "%g %g lineto\n", vertex[i].x, vertex[i].y);
        }
        fprintf(file, "closepath fill\n\n");
      }
    }
    loc += nvertices * 7;  /* Each vertex element in the
                              feedback buffer is 7 GLfloats. */
    break;
  case GL_POINT_TOKEN:
    vertex = (Feedback3Dcolor *) loc;
    fprintf(file, "%g %g %g setrgbcolor\n", vertex[0].red, vertex[0].green, vertex[0].blue);
    fprintf(file, "%g %g %g 0 360 arc fill\n\n", vertex[0].x, vertex[0].y, pointSize / 2.0);
    loc += 7;           /* Each vertex element in the feedback
                           buffer is 7 GLfloats. */
    break;
  default:
    /* XXX Left as an excersie to the reader. */
    printf("Incomplete implementation.  Unexpected token (%d).\n", token);
    exit(1);
  }
  return loc;
}

void
CKPlot3D::spewUnsortedFeedback(FILE * file, GLint size, GLfloat * buffer)
{
  GLfloat *loc, *end;

  loc = buffer;
  end = buffer + size;
  while (loc < end) {
    loc = spewPrimitiveEPS(file, loc);
  }
}

typedef struct _DepthIndex {
  GLfloat *ptr;
  GLfloat depth;
} DepthIndex;

static int
compare(const void *a, const void *b)
{
  DepthIndex *p1 = (DepthIndex *) a;
  DepthIndex *p2 = (DepthIndex *) b;
  GLfloat diff = p2->depth - p1->depth;

  if (diff > 0.0) {
    return 1;
  } else if (diff < 0.0) {
    return -1;
  } else {
    return 0;
  }
}

void
CKPlot3D::spewSortedFeedback(FILE * file, GLint size, GLfloat * buffer)
{
  int token;
  GLfloat *loc, *end;
  Feedback3Dcolor *vertex;
  GLfloat depthSum;
  int nprimitives, item;
  DepthIndex *prims;
  int nvertices, i;

  end = buffer + size;

  /* Count how many primitives there are. */
  nprimitives = 0;
  loc = buffer;
  while (loc < end) {
    token = *loc;
    loc++;
    switch (token) {
    case GL_LINE_TOKEN:
    case GL_LINE_RESET_TOKEN:
      loc += 14;
      nprimitives++;
      break;
    case GL_POLYGON_TOKEN:
      nvertices = *loc;
      loc++;
      loc += (7 * nvertices);
      nprimitives++;
      break;
    case GL_POINT_TOKEN:
      loc += 7;
      nprimitives++;
      break;
    default:
      /* XXX Left as an excersie to the reader. */
      printf("Incomplete implementation.  Unexpected token (%d).\n",
        token);
      exit(1);
    }
  }

  /* Allocate an array of pointers that will point back at
     primitives in the feedback buffer.  There will be one
     entry per primitive.  This array is also where we keep the
     primitive's average depth.  There is one entry per
     primitive  in the feedback buffer. */
  prims = (DepthIndex *) malloc(sizeof(DepthIndex) * nprimitives);

  item = 0;
  loc = buffer;
  while (loc < end) {
    prims[item].ptr = loc;  /* Save this primitive's location. */
    token = *loc;
    loc++;
    switch (token) {
    case GL_LINE_TOKEN:
    case GL_LINE_RESET_TOKEN:
      vertex = (Feedback3Dcolor *) loc;
      depthSum = vertex[0].z + vertex[1].z;
      prims[item].depth = depthSum / 2.0;
      loc += 14;
      break;
    case GL_POLYGON_TOKEN:
      nvertices = *loc;
      loc++;
      vertex = (Feedback3Dcolor *) loc;
      depthSum = vertex[0].z;
      for (i = 1; i < nvertices; i++) {
        depthSum += vertex[i].z;
      }
      prims[item].depth = depthSum / nvertices;
      loc += (7 * nvertices);
      break;
    case GL_POINT_TOKEN:
      vertex = (Feedback3Dcolor *) loc;
      prims[item].depth = vertex[0].z;
      loc += 7;
      break;
    default:
      /* XXX Left as an excersie to the reader. */
      assert(1);
    }
    item++;
  }
  assert(item == nprimitives);

  /* Sort the primitives back to front. */
  qsort(prims, nprimitives, sizeof(DepthIndex), compare);

  /* Understand that sorting by a primitives average depth
     doesn't allow us to disambiguate some cases like self
     intersecting polygons.  Handling these cases would require
     breaking up the primitives.  That's too involved for this
     example.  Sorting by depth is good enough for lots of
     applications. */

  /* Emit the Encapsulated PostScript for the primitives in
     back to front order. */
  for (item = 0; item < nprimitives; item++) {
    (void) spewPrimitiveEPS(file, prims[item].ptr);
  }

  free(prims);
}

#define EPS_GOURAUD_THRESHOLD 0.1  /* Lower for better (slower) 

                                      smooth shading. */

void
CKPlot3D::spewWireFrameEPS(FILE * file, int doSort, GLint size, GLfloat * buffer, char *creator)
{
  GLfloat clearColor[4], viewport[4];
  GLfloat lineWidth;
  int i;

  /* Read back a bunch of OpenGL state to help make the EPS
     consistent with the OpenGL clear color, line width, point
     size, and viewport. */
  glGetFloatv(GL_VIEWPORT, viewport);
  glGetFloatv(GL_COLOR_CLEAR_VALUE, clearColor);
  glGetFloatv(GL_LINE_WIDTH, &lineWidth);
  glGetFloatv(GL_POINT_SIZE, &pointSize);

  /* Emit EPS header. */
  fputs("%!PS-Adobe-2.0 EPSF-2.0\n", file);
  /* Notice %% for a single % in the fprintf calls. */
  fprintf(file, "%%%%Creator: %s (using OpenGL feedback)\n", file, creator);
  fprintf(file, "%%%%BoundingBox: %g %g %g %g\n",
    viewport[0], viewport[1], viewport[2], viewport[3]);
  fputs("%%EndComments\n", file);
  fputs("\n", file);
  fputs("gsave\n", file);
  fputs("\n", file);

  /* Output Frederic Delhoume's "gouraudtriangle" PostScript
     fragment. */
  fputs("% the gouraudtriangle PostScript fragement below is free\n", file);
  fputs("% written by Frederic Delhoume (delhoume@ilog.fr)\n", file);
  fprintf(file, "/threshold %g def\n", EPS_GOURAUD_THRESHOLD);
  for (i = 0; gouraudtriangleEPS[i]; i++) {
    fprintf(file, "%s\n", gouraudtriangleEPS[i]);
  }

  fprintf(file, "\n%g setlinewidth\n", lineWidth);

  /* Clear the background like OpenGL had it. */
  fprintf(file, "%g %g %g setrgbcolor\n",
    clearColor[0], clearColor[1], clearColor[2]);
  fprintf(file, "%g %g %g %g rectfill\n\n",
    viewport[0], viewport[1], viewport[2], viewport[3]);

  if (doSort) {
    spewSortedFeedback(file, size, buffer);
  } else {
    spewUnsortedFeedback(file, size, buffer);
  }

  /* Emit EPS trailer. */
  fputs("grestore\n\n", file);
  fputs("%Add `showpage' to the end of this file to be able to print to a printer.\n",
    file);

  fclose(file);
}

void
CKPlot3D::outputEPS(int size, int doSort, char *filename)
{
  GLfloat *feedbackBuffer;
  GLint returned;
  FILE *file;

  feedbackBuffer = (GLfloat*)calloc(size, sizeof(GLfloat));
  glFeedbackBuffer(size, GL_3D_COLOR, feedbackBuffer);
  (void) glRenderMode(GL_FEEDBACK);
  draw();
  returned = glRenderMode(GL_RENDER);
  if (filename) {
    file = fopen(filename, "w");
    if (file) {
      spewWireFrameEPS(file, doSort, returned, feedbackBuffer, "rendereps");
    } else {
      printf("Could not open %s\n", filename);
    }
  } else {
    /* Helps debugging to be able to see the decode feedback
       buffer as text. */
    printBuffer(returned, feedbackBuffer);
  }
  free(feedbackBuffer);
}


void CKPlot3D::RenderPixmap(int n)
{
  aktRenderPixmap=n;
}


void CKPlot3D::processDrawCommand(float *ar , int *funcIn)
{
   char buf[81];
   float ar1,ar2,ar3,ar4;
    int func;
    int lwert;
   float t_x,t_y;
   int j;
   func=*funcIn;
  
  switch(func)
  {
    case 0:
        EndDraw();
		break;
    case 6:
        BeginDraw();
        break;
    case 33:
        zPosOrg=(ar[0]);
        zPos=zPosOrg;
//        fprintf(stderr,"33 %f\n",zPosOrg);
       	break;
    case 34:
		{
			 int nr=*(( int*)(&(ar[0])));
			float r=(ar[1]);
			 int color=*(( int*)(&(ar[4])));
			SetAtom3DDisp(nr, r, color); 
//			fprintf(stderr,"34 %d %f %d\n",nr,r,color);
		}
		break;
    case 35:
		{
			 int nr1=*(( int*)(&(ar[0])));
			 int nr2=*(( int*)(&(ar[1])));
			float r=(ar[2]);
			 int color=*(( int*)(&(ar[6])));
			SetBond3D(nr1,nr2,color,r);
//			fprintf(stderr,"35 %d %d %f %d\n",nr1,nr2,r,color);
		}
    break;
    case 36:
		{
			float x=(ar[0]);
			float y=(ar[1]);
			float z=(ar[2]);
			 int nr=*(( int*)(&(ar[3])));
			SetAtom3D(x,y,z,nr);         
//			fprintf(stderr,"36 %f %f %f %d\n",x,y,z,nr);
		}
    break;
    case 37:
		{
			Polygon3D *p=new Polygon3D();
			 int nr=*(( int*)(&(ar[0])));
			 int color=*(( int*)(&(ar[1])));
			 int muster=*(( int*)(&(ar[2])));
			p->color=color;
			p->muster=muster;
			p->glDisplayList=0;
//			fprintf(stderr,"37 %d %d %d",nr,color,muster);
			int i=nr-3;
			for (int j=0;j<i;j++)
			{
				 int nrAtom=*(( int*)(&(ar[j+3])));
				int *n=new int();
				*n=nrAtom;
				p->atome.append(n);
//				fprintf(stderr," %d",nrAtom);
			}	
			polygonList.append(p);
//			fprintf(stderr,"\n");
		}
    default:
//	  lwert=*((long int*)(&(ar[0])));
	   ;
 }
}

